def enable() -> bool:
    ...

def disable() -> bool:
    ...

def info(f: callable) -> dict:
    ...

def dump_il(f: callable) -> bytearray:
    ...

def dump_native(f: callable) -> bytearray:
    ...
